@@ -1,9 +1,8 @@
 """
-Description: Test the validator
+Description: Test stac-validator on item-collection validation.
 
 """
 
-
 from stac_validator import stac_validator
 
 
